/***************************************************************************/
/*                      THintBox version 0.65                              */
/* ----------------------------------------------------------------------- */
/* What is it: This is a quick implementation of the HintBox's that you can*/
/* ----------- see in many Lotus products (for instance Lotus SmartSuite). */
/*                                                                         */
/* How to install it: Go to the "Component" Menuitem and select "Install   */
/* ------------------ Component...". Click the "Browse" button next to the */
/*                    "Unit file name:". Choose "HintBox.cpp".             */
/*                    Choose the apporpriate package to install THintBox in*/
/*                    to, from the "Package file name" combobox (normally: */
/*                    dclusr35.bpk or dclusr40.bpk)                        */
/*                    THintBox will register it self in the "Samples"      */
/*                    palette.                                             */
/*                                                                         */
/* How to use it: Just drop it on to the Form you're working with and work */
/* -------------- on.                                                      */
/*                                                                         */
/* What does it cost: Nothing. It's free of charge for any usage! :-)      */
/* ------------------                                                      */
/*                                                                         */
/* Which version of C++ Builder is it tested with: 3 and 4. Should work    */
/* ----------------------------------------------- with version 1 too.     */
/*                                                                         */
/* Disclaimer:                                                             */
/* -----------                                                             */
/* THintBox is provided "AS IS" without warranty of any                    */
/* kind, either expressed or implied, including but not limited            */
/* to the implied warranties of merchantability and fitness                */
/* for a particular purpose. In no event shall Cramon Utilites             */
/* be liable for any damages whatsoever including direct, indirect,        */
/* incidental, consequential, loss of business profits or                  */
/* special damages, even if Cramon Utilities has been advised              */
/* of the possibility of such damages.                                     */
/* You may use THintBox free of charge in any application you make         */
/* be it freeware, shareware or commercial. If you change the sourcecode   */
/* of THintBox you must include this text and clearly state which parts    */
/* are changed by you and that this is not the original version.           */
/* If you would like your changes in the "official" version please send    */
/* the updated source file(s) to me and I'll update my sourcecode (you'll  */
/* ofcourse get credit for all changes you've made to THintBox.            */
/*                                                                         */
/* Author..: Jeppe Cramon                                                  */
/* Company.: Cramon Utilities                                              */
/* WWW.....: http://www.cramon.dk                                          */
/* E-mail..: jeppe@cramon.dk                                               */
/* (C) Copyright 1999 Cramon Utilities                                     */
/*                                                                         */
/* ----------------------------------------------------------------------- */
/* Date         Comment                                 Programmer         */
/* 05/03-1999   Created the first version (0.20)        Jeppe Cramon       */
/* 05/04-1999   Version 0.21. Fixed a problem with the  Jeppe Cramon       */
/*              space getting drawn on the next line                       */
/*              when it should have been removed (as a                     */
/*              part of the WordWrap)                                      */
/* 05/08-1999   Changed inheritance from TPanel, to     Jeppe Cramon       */
/*              TCustomControl. Changed from Caption to                    */
/*              Lines, just like in a TMemo, so now you                    */
/*              can have multiple lines with WordWrap                      */
/*              instead only one.                                          */
/*              NEW: You can now set ShadowSize.                           */
/* 05/09-1999   Add support for HTML tags. This means   Jeppe Cramon       */
/*              that THintBox, now is able to display                      */
/*              HTML formattet text. So far only a very                    */
/*              minor fraction of the HTML standard is                     */
/*              supported:                                                 */
/*               Center Text - <CENTER>..Text..</CENTER>                   */
/*               Line break  - <BR>                                        */
/*               Bold        - <B>..Text..</B>                             */
/*               Italic      - <I>..Text..</I>                             */
/*               Underline   - <U>..Text..</U>                             */
/*               StrikeOut   - <STRIKE>..Text..</STRIKE>                   */
/*              NEW: You can now also choose if THintBox                   */
/*                   should draw it's hintbox. Setting                     */
/*                   the DrawHintBox property to false                     */
/*                   will make it look like a label, but                   */
/*                   with all the HintBox features ;-)                     */
/***************************************************************************/

//---------------------------------------------------------------------------
#include <vcl.h>
#pragma hdrstop

#include "HintBox.h"
#pragma package(smart_init)
//---------------------------------------------------------------------------
// ValidCtrCheck is used to assure that the components created do not have
// any pure virtual functions.
//
static inline void ValidCtrCheck(THintBox *)
{
    new THintBox(NULL);
}
//---------------------------------------------------------------------------
__fastcall THintBox::THintBox(TComponent* Owner)
    : TCustomControl(Owner)
{
    // Set default property values
    BkColor = clInfoBk;
    ShColor = clGray;
    iShadowSize = 4;
    bHTMLFormatting = true;
    DoubleBuffered = true;
    bDrawBox = true;

    FLines = new TStringList();
    TextList = NULL;
    CenterList = NULL;
}
//---------------------------------------------------------------------------
__fastcall THintBox::~THintBox()
{
    if (FLines)
        delete FLines;
    if (TextList)
        delete TextList;
    // To do (proper clean up)
    if (CenterList)
        delete CenterList;
}
//---------------------------------------------------------------------------
namespace Hintbox
{
    void __fastcall PACKAGE Register()
    {
         TComponentClass classes[1] = {__classid(THintBox)};
         RegisterComponents("Samples", classes, 0);
    }
}
//---------------------------------------------------------------------------
void __fastcall THintBox::DrawBox()
{
    TRect Rect = ClientRect;

    if (bDrawBox)
    {
        Canvas->Brush->Style = bsSolid;

        // Draw shadow
        Rect.Top += iShadowSize;
        Rect.Left += iShadowSize;
        Canvas->Brush->Color = ShColor;
        Canvas->FillRect(Rect);

        // Draw hint frame
        Rect = ClientRect;
        Rect.Right -= iShadowSize;
        Rect.Bottom -= iShadowSize;
        Canvas->Brush->Color = BkColor;
        Canvas->Font->Color = clBlack;
        Canvas->Rectangle(Rect.Left, Rect.Top, Rect.Right, Rect.Bottom);
        Canvas->Font->Color = Font->Color;
    }
    else
    {
        Canvas->Font->Color = Font->Color;
        Canvas->Brush->Style = bsClear;
        Invalidate();
    }


    // Stop painting if there's no text available
    if (FLines->Count == 0)
        return;

    /******************************************/
    /* --- Write the text (with WordWrap) --- */
    /******************************************/

    int iLastSpace=0;                       // Index of the last encountered space
    int iPos1 = 1, iPos2 = 1, iLineNo=0;    // Startindex, endindex and number of lines drawn
    TSize size;                             // Size (used in TextExtent)
    AnsiString asTmp;                       // Tmp. string (used in TextExtent)
    AnsiString Caption;                     // Used for quick porting of the old component, that used TPanel's Caption.

    // Set start coordinates
    if (bDrawBox)
    {
        Rect.Left += 4;
        Rect.Top += 4;
    }

    // Loop through all the lines
    for (int iLoop = 0; iLoop < FLines->Count; iLoop++)
    {
        // Init this run
        Caption = FLines->Strings[iLoop];
        iPos1 = 1, iPos2 = 1;

        // Skip the line if it's empty
        if (Caption == "")
            goto CONTINUE;

        do
        {
            iPos2++;
            asTmp = Caption.SubString(iPos1, iPos2-iPos1);
            size = Canvas->TextExtent(asTmp);
            //size = TextExtentEx(asTmp);

            if (asTmp[asTmp.Length()] == ' ')
                iLastSpace = iPos2-1;

            if (size.cx > Rect.Right - Rect.Left-iShadowSize)
            {
                // Check if the iLastSpace is zero (otherwise no text at all would be written)
                if (iLastSpace==0)
                    iLastSpace = iPos2-1;

                // Make sure it doesn't extend the frames bottom
                if (Rect.Top+size.cy*(iLineNo+1) >= Rect.Bottom)
                    return;

                // Write string to box
                Canvas->TextOut(Rect.Left, Rect.Top+size.cy*iLineNo, Caption.SubString(iPos1, iLastSpace-iPos1));
                //TextOutEx(Rect.Left, Rect.Top+size.cy*iLineNo, Caption.SubString(iPos1, iLastSpace-iPos1));
                iPos2 = iLastSpace+1;
                iLastSpace = 0;
                iPos1 = iPos2;
                iLineNo++;
            }
        } while(iPos2 <= Caption.Length());

        // Write the rest of the string (or the entire string, incase there was no WordWrapping required)
        // Make sure we don't extend the bottom of the box
        if (Rect.Top+size.cy*(iLineNo+1) < Rect.Bottom)
        {
            //TextOutEx(Rect.Left, Rect.Top+size.cy*iLineNo, Caption.SubString(iPos1, iPos2-iPos1+1));
            Canvas->TextOut(Rect.Left, Rect.Top+size.cy*iLineNo, Caption.SubString(iPos1, iPos2-iPos1+1));
        }

CONTINUE:
        iLineNo++;      // Next line
    } // end for
}
//---------------------------------------------------------------------------
void __fastcall THintBox::DrawBoxEx()
{
    static TFontStyles style = Canvas->Font->Style;

    // Check if we have performed the first parse
    if (!TextList)
    {
        TextList = new TStringList();
        CenterList = new TList();
        Parse();
    }

    // Reset font style
    Canvas->Font->Style = style;

    // Get size of Rectangle
    TRect Rect = ClientRect;

    if (bDrawBox)
    {
        Canvas->Brush->Style = bsSolid;

        // Draw shadow
        Rect.Top += iShadowSize;
        Rect.Left += iShadowSize;
        Canvas->Brush->Color = ShColor;
        Canvas->FillRect(Rect);

        // Draw hint frame
        Rect = ClientRect;
        Rect.Right -= iShadowSize;
        Rect.Bottom -= iShadowSize;
        Canvas->Brush->Color = BkColor;
        Canvas->Font->Color = clBlack;
        Canvas->Rectangle(Rect.Left, Rect.Top, Rect.Right, Rect.Bottom);
        Canvas->Font->Color = Font->Color;
    }
    else
    {
        Canvas->Font->Color = Font->Color;
        Canvas->Brush->Style = bsClear;
        Invalidate();
    }


    // Stop painting if there's no text available
    if (TextList->Count == 0)
        return;

    /******************************************/
    /* --- Write the text (with WordWrap) --- */
    /******************************************/

    int iLastSpace=0;                       // Index of the last encountered space
    int iPos1 = 1, iPos2 = 1, iLineNo=0;    // Startindex, endindex and number of lines drawn
    TSize size;                             // Size (used in TextExtent)
    AnsiString asTmp;                       // Tmp. string (used in TextExtent)
    AnsiString Caption;                     // Used for quick porting of the old component, that used TPanel's Caption.

    // Set start coordinates
    if (bDrawBox)
    {
        Rect.Left += 4;
        Rect.Top += 4;
    }

    Canvas->Font->Style = Canvas->Font->Style.Clear();

    // Loop through all the lines
    for (int iLoop = 0; iLoop < TextList->Count; iLoop++)
    {
        // Init this run
        Caption = TextList->Strings[iLoop];
        iPos1 = 1, iPos2 = 1;
        size = TextExtentEx(Caption);

        // Skip the line if it's empty
        if (Caption == "")
            goto CONTINUE;
        else
        if (size.cx > Rect.Right - Rect.Left-iShadowSize)
        {
            do
            {
                iPos2++;
                asTmp = Caption.SubString(iPos1, iPos2-iPos1);
                //size = Canvas->TextExtent(asTmp);
                size = TextExtentEx(asTmp);

                if (asTmp[asTmp.Length()] == ' ')
                    iLastSpace = iPos2-1;

                if (size.cx > Rect.Right - Rect.Left-iShadowSize)
                {
                    // Check if the iLastSpace is zero (otherwise no text at all would be written)
                    if (iLastSpace==0)
                            iLastSpace = iPos2-1;

                    // Make sure it doesn't extend the frames bottom
                    if (Rect.Top+size.cy*(iLineNo+1) >= Rect.Bottom)
                            return;

                    // Write string to box
                    //Canvas->TextOut(Rect.Left, Rect.Top+size.cy*iLineNo, Caption.SubString(iPos1, iLastSpace-iPos1));
                    TextOutEx(Rect.Left, Rect.Top+size.cy*iLineNo, Caption.SubString(iPos1, iLastSpace-iPos1));
                    iPos2 = iLastSpace+1;
                    iLastSpace = 0;
                    iPos1 = iPos2;
                    iLineNo++;
                } // end if
            } while(iPos2 <= Caption.Length());
        }
        else
        {
            iPos1 = 1;
            iPos2 = Caption.Length();
        } // end if

        // Write the rest of the string (or the entire string, incase there was no WordWrapping required)
        // Make sure we don't extend the bottom of the box
        if (Rect.Top+size.cy*(iLineNo+1) < Rect.Bottom)
        {
            if (CenterList->Items[ iLoop ])
            {
                //Canvas->TextOut(Rect.Left + (Rect.Right-Rect.Left-size.cx)/2-4, Rect.Top+size.cy*iLineNo, Caption.SubString(iPos1, iPos2-iPos1+1));
                TextOutEx(Rect.Left + (Rect.Right-Rect.Left-size.cx)/2-4, Rect.Top+size.cy*iLineNo, Caption.SubString(iPos1, iPos2-iPos1+1));
            }
            else
                TextOutEx(Rect.Left, Rect.Top+size.cy*iLineNo, Caption.SubString(iPos1, iPos2-iPos1+1));
                //Canvas->TextOut(Rect.Left, Rect.Top+size.cy*iLineNo, Caption.SubString(iPos1, iPos2-iPos1+1));
        }

CONTINUE:
        iLineNo++;      // Next line
    } // end for
}
//---------------------------------------------------------------------------
void __fastcall THintBox::Paint()
{
    if (bHTMLFormatting)
        DrawBoxEx();
    else
        DrawBox();
}
//---------------------------------------------------------------------------
void __fastcall THintBox::SetColor(TColor Value)
{
    if (Value != BkColor)
    {
        BkColor = Value;
        Paint();
    }
}
//---------------------------------------------------------------------------
void __fastcall THintBox::SetShadowColor(TColor Value)
{
    if (Value != ShColor)
    {
        ShColor = Value;
        Paint();
    }
}
//---------------------------------------------------------------------------
void __fastcall THintBox::SetLines(Classes::TStrings* Value)
{
    FLines->Assign(Value);

    if (bHTMLFormatting)
        Parse();

    Paint();
}
//---------------------------------------------------------------------------
void __fastcall THintBox::SetShadowSize(int Value)
{
    if (Value != iShadowSize && Value >= 0)
    {
        Width += Value-iShadowSize;
        Height += Value-iShadowSize;
        iShadowSize = Value;
        Paint();
    }
}
//---------------------------------------------------------------------------
void __fastcall THintBox::Parse()
{
    // Clean up
    if (!TextList)
    {
        TextList = new TStringList();
        CenterList = new TList();
    }
    else
    {
        CenterList->Clear();
        TextList->Clear();
    }

    // Check if FLines is empty
    if (FLines->Count == 0)
        return;

    // Parse the contents of FLines
    for (int iLoop=0; iLoop < FLines->Count; iLoop++)
    {
        ParseLine( FLines->Strings[ iLoop ] );
    } // end for
}
//---------------------------------------------------------------------------
int __fastcall THintBox::FindNext(AnsiString asText, AnsiString asFind, int iOldIndex)
{
    AnsiString asTmp;
    int iIndex = iOldIndex;
    while(iIndex <= asText.Length())
    {
        asTmp = asText.SubString(iIndex, asFind.Length());
        if (asTmp == asFind)
            return iIndex;
        else
            iIndex++;
    }
    return 0;
}
//---------------------------------------------------------------------------
void __fastcall THintBox::ParseLine(AnsiString asText)
{
    // Local variables
    int iLastPos = 1;
    int iEndPos;
    int iStartPos;
    bool bCenter = false;

    while(1)
    {
        // Start parsing the line
        // Check if it contains a HTML tag
        // Find the start of the tag
SEARCH_AGAIN:
        if (asText == "")
        {
            TextList->Add(asText);
            CenterList->Add((void*)false);
            return;
        }


        iStartPos = FindNext(asText,"<", iLastPos);
        if (iStartPos == 0)
        {
            // It didn't contain a start of tag
            TextList->Add(asText);
            CenterList->Add((void*)false);
            return;
        }

        // Check if it an end of tag (contains ">" without the ("))
        iEndPos = FindNext(asText,">", iLastPos);
        if (iEndPos == 0)
        {
            TextList->Add(asText);
            CenterList->Add((void*)false);
            return;
        }
        else if (iEndPos < iStartPos)
        {
            iLastPos = iEndPos+1;
            goto SEARCH_AGAIN;
        }

        AnsiString asTmp = asText.SubString(iStartPos+1, iEndPos-iStartPos-1).UpperCase();
        if ( asTmp == "CENTER")
        {
            if (iStartPos != 1)
            {
                // We're starting on a new line, so move the start of the string to a line for it self
                TextList->Add(asText.SubString(1, iStartPos-1));
            }
            iLastPos = 1;
            asText = asText.SubString(iEndPos+1, asText.Length()-iEndPos).TrimLeft();
            CenterList->Add((void*)true);
            bCenter = true;
        }
        else if (asTmp == "BR")
        {
            TextList->Add(asText.SubString(1, iStartPos-1));
            CenterList->Add((void*)bCenter);
            asText = asText.SubString(iEndPos+1, asText.Length()-iEndPos).TrimLeft();
            iLastPos = 1;
        }
        else if ( asTmp == "/CENTER")
        {
            TextList->Add( asText.SubString(1, iStartPos-1));
            CenterList->Add((void*)false);
            asText = asText.SubString(iEndPos+1, asText.Length()-iEndPos).TrimLeft();
            iLastPos = 1;
            bCenter = false;
            if (asText == "")
                return;
        }
        else
            iLastPos = iStartPos+1;

    } // end while(1)
}
//---------------------------------------------------------------------------
TSize __fastcall THintBox::TextExtentEx(AnsiString asTestText)
{
    // Local variables
    TSize tmpSize;
    TFontStyles style = Canvas->Font->Style;
    AnsiString asText = asTestText;
    int iLastPos = 1;
    int iEndPos;
    int iStartPos;


    /* --- Init --- */
    tmpSize.cx = 0;
    tmpSize.cy = 0;

    while(1)
    {
        // Start parsing the line
        // Check if it contains a HTML tag
        // Find the start of the tag
SEARCH_AGAIN:
        if (asText == "")
        {
            // Reset style
            Canvas->Font->Style = style;

            return tmpSize;
        }

        iStartPos = FindNext(asText,"<", iLastPos);
        if (iStartPos == 0)
        {
            // It didn't contain a start of tag
            TSize size = Canvas->TextExtent(asText);
            tmpSize.cx += size.cx;

            // Check if we need to update cy
            if (size.cy > tmpSize.cy)
                tmpSize.cy = size.cy;

            // Reset style
            Canvas->Font->Style = style;

            return tmpSize;
        }

        // Check if it an end of tag (contains ">" without the ("))
        iEndPos = FindNext(asText,">", iLastPos);
        if (iEndPos == 0)
        {
            // It didn't contain an end of tag
            TSize size = Canvas->TextExtent(asText);
            tmpSize.cx += size.cx;

            // Check if we need to update cy
            if (size.cy > tmpSize.cy)
                tmpSize.cy = size.cy;

            // Reset style
            Canvas->Font->Style = style;

            return tmpSize;
        }
        else if (iEndPos < iStartPos)
        {
            iLastPos = iEndPos+1;
            goto SEARCH_AGAIN;
        }

        AnsiString asTmp = asText.SubString(iStartPos+1, iEndPos-iStartPos-1).UpperCase();
        if ( asTmp == "B")
        {
            // Get extents of current text
            TSize size = Canvas->TextExtent(asText.SubString(1, iStartPos-1));
            tmpSize.cx += size.cx;

            // Check if we need to update cy
            if (size.cy > tmpSize.cy)
                tmpSize.cy = size.cy;

            // Set new font
            Canvas->Font->Style = Canvas->Font->Style << fsBold;
            // Skip the <..> part
            asText = asText.SubString(iEndPos+1, asText.Length()-iEndPos);
            iLastPos = 1;
        }
        else if ( asTmp == "I")
        {
            // Get extents of current text
            TSize size = Canvas->TextExtent(asText.SubString(1, iStartPos-1));
            tmpSize.cx += size.cx;

            // Check if we need to update cy
            if (size.cy > tmpSize.cy)
                tmpSize.cy = size.cy;

            // Set new font
            Canvas->Font->Style = Canvas->Font->Style << fsItalic;
            // Skip the <..> part
            asText = asText.SubString(iEndPos+1, asText.Length()-iEndPos);
            iLastPos = 1;

        }
        else if ( asTmp == "U")
        {
            // Get extents of current text
            TSize size = Canvas->TextExtent(asText.SubString(1, iStartPos-1));
            tmpSize.cx += size.cx;

            // Check if we need to update cy
            if (size.cy > tmpSize.cy)
                tmpSize.cy = size.cy;

            // Set new font
            Canvas->Font->Style = Canvas->Font->Style << fsUnderline;
            // Skip the <..> part
            asText = asText.SubString(iEndPos+1, asText.Length()-iEndPos);
            iLastPos = 1;

        }
        else if ( asTmp == "STRIKE")
        {
            // Get extents of current text
            TSize size = Canvas->TextExtent(asText.SubString(1, iStartPos-1));
            tmpSize.cx += size.cx;

            // Check if we need to update cy
            if (size.cy > tmpSize.cy)
                tmpSize.cy = size.cy;

            // Set new font
            Canvas->Font->Style = Canvas->Font->Style << fsStrikeOut;
            // Skip the <..> part
            asText = asText.SubString(iEndPos+1, asText.Length()-iEndPos);
            iLastPos = 1;

        }
        else if ( asTmp == "/B")
        {
            // Get extents of current text
            TSize size = Canvas->TextExtent(asText.SubString(1, iStartPos-1));
            tmpSize.cx += size.cx;

            // Check if we need to update cy
            if (size.cy > tmpSize.cy)
                tmpSize.cy = size.cy;

            // Set new font
            Canvas->Font->Style = Canvas->Font->Style >> fsBold;
            // Skip the <..> part
            asText = asText.SubString(iEndPos+1, asText.Length()-iEndPos);
            iLastPos = 1;
        }
        else if ( asTmp == "/I")
        {
            // Get extents of current text
            TSize size = Canvas->TextExtent(asText.SubString(1, iStartPos-1));
            tmpSize.cx += size.cx;

            // Check if we need to update cy
            if (size.cy > tmpSize.cy)
                tmpSize.cy = size.cy;

            // Set new font
            Canvas->Font->Style = Canvas->Font->Style >> fsItalic;
            // Skip the <..> part
            asText = asText.SubString(iEndPos+1, asText.Length()-iEndPos);
            iLastPos = 1;

        }
        else if ( asTmp == "/U")
        {
            // Get extents of current text
            TSize size = Canvas->TextExtent(asText.SubString(1, iStartPos-1));
            tmpSize.cx += size.cx;

            // Check if we need to update cy
            if (size.cy > tmpSize.cy)
                tmpSize.cy = size.cy;

            // Set new font
            Canvas->Font->Style = Canvas->Font->Style >> fsUnderline;
            // Skip the <..> part
            asText = asText.SubString(iEndPos+1, asText.Length()-iEndPos);
            iLastPos = 1;

        }
        else if ( asTmp == "/STRIKE")
        {
            // Get extents of current text
            TSize size = Canvas->TextExtent(asText.SubString(1, iStartPos-1));
            tmpSize.cx += size.cx;

            // Check if we need to update cy
            if (size.cy > tmpSize.cy)
                tmpSize.cy = size.cy;

            // Set new font
            Canvas->Font->Style = Canvas->Font->Style >> fsStrikeOut;
            // Skip the <..> part
            asText = asText.SubString(iEndPos+1, asText.Length()-iEndPos);
            iLastPos = 1;

        }
        else
        {
           // Unsupported tag
            // Get extents of current text
            TSize size = Canvas->TextExtent(asText.SubString(1, iEndPos));
            tmpSize.cx += size.cx;
            asText = asText.SubString(iEndPos+1, asText.Length()-iEndPos);

            // Check if we need to update cy
            if (size.cy > tmpSize.cy)
                tmpSize.cy = size.cy;

            iLastPos = 1;
//            iLastPos = iStartPos+1;
        }
    } // end while(1)
}
//---------------------------------------------------------------------------
void __fastcall THintBox::TextOutEx(int X, int Y, const AnsiString asTextOut)
{
    // Local variables
    AnsiString asText = asTextOut;
    TSize tmpSize; tmpSize.cx = 0;
    int iLastPos = 1;
    int iEndPos;
    int iStartPos;

    while(1)
    {
        // Start parsing the line
        // Check if it contains a HTML tag
        // Find the start of the tag
SEARCH_AGAIN:
        if (asText == "")
        {
            return;
        }

        iStartPos = FindNext(asText,"<", iLastPos);
        if (iStartPos == 0)
        {
            Canvas->TextOut(tmpSize.cx+X, Y, asText);
            return;
        }

        // Check if it an end of tag (contains ">" without the ("))
        iEndPos = FindNext(asText,">", iLastPos);
        if (iEndPos == 0)
        {
            Canvas->TextOut(tmpSize.cx+X, Y, asText);
            return;
        }
        else if (iEndPos < iStartPos)
        {
            iLastPos = iEndPos+1;
            goto SEARCH_AGAIN;
        }

        AnsiString asTmp = asText.SubString(iStartPos+1, iEndPos-iStartPos-1).UpperCase();
        if ( asTmp == "B")
        {
            Canvas->TextOut(tmpSize.cx+X, Y, asText.SubString(1, iStartPos-1));
            tmpSize.cx += Canvas->TextExtent(asText.SubString(1, iStartPos-1)).cx;

            // Set new font
            Canvas->Font->Style = Canvas->Font->Style << fsBold;
            // Skip the <..> part
            asText = asText.SubString(iEndPos+1, asText.Length()-iEndPos);
            iLastPos = 1;
        }
        else if ( asTmp == "I")
        {
            Canvas->TextOut(tmpSize.cx+X, Y, asText.SubString(1, iStartPos-1));
            tmpSize.cx += Canvas->TextExtent(asText.SubString(1, iStartPos-1)).cx;


            // Set new font
            Canvas->Font->Style = Canvas->Font->Style << fsItalic;
            // Skip the <..> part
            asText = asText.SubString(iEndPos+1, asText.Length()-iEndPos);
            iLastPos = 1;

        }
        else if ( asTmp == "U")
        {
            Canvas->TextOut(tmpSize.cx+X, Y, asText.SubString(1, iStartPos-1));
            tmpSize.cx += Canvas->TextExtent(asText.SubString(1, iStartPos-1)).cx;


            // Set new font
            Canvas->Font->Style = Canvas->Font->Style << fsUnderline;
            // Skip the <..> part
            asText = asText.SubString(iEndPos+1, asText.Length()-iEndPos);
            iLastPos = 1;

        }
        else if ( asTmp == "STRIKE")
        {
            Canvas->TextOut(tmpSize.cx+X, Y, asText.SubString(1, iStartPos-1));
            tmpSize.cx += Canvas->TextExtent(asText.SubString(1, iStartPos-1)).cx;

            // Set new font
            Canvas->Font->Style = Canvas->Font->Style << fsStrikeOut;
            // Skip the <..> part
            asText = asText.SubString(iEndPos+1, asText.Length()-iEndPos);
            iLastPos = 1;

        }
        else if ( asTmp == "/B")
        {
            Canvas->TextOut(tmpSize.cx+X, Y, asText.SubString(1, iStartPos-1));
            tmpSize.cx += Canvas->TextExtent(asText.SubString(1, iStartPos-1)).cx;

            // Set new font
            Canvas->Font->Style = Canvas->Font->Style >> fsBold;
            // Skip the <..> part
            asText = asText.SubString(iEndPos+1, asText.Length()-iEndPos);
            iLastPos = 1;
        }
        else if ( asTmp == "/I")
        {
            Canvas->TextOut(tmpSize.cx+X, Y, asText.SubString(1, iStartPos-1));
            tmpSize.cx += Canvas->TextExtent(asText.SubString(1, iStartPos-1)).cx;

            // Set new font
            Canvas->Font->Style = Canvas->Font->Style >> fsItalic;
            // Skip the <..> part
            asText = asText.SubString(iEndPos+1, asText.Length()-iEndPos);
            iLastPos = 1;

        }
        else if ( asTmp == "/U")
        {
            Canvas->TextOut(tmpSize.cx+X, Y, asText.SubString(1, iStartPos-1));
            tmpSize.cx += Canvas->TextExtent(asText.SubString(1, iStartPos-1)).cx;

            // Set new font
            Canvas->Font->Style = Canvas->Font->Style >> fsUnderline;
            // Skip the <..> part
            asText = asText.SubString(iEndPos+1, asText.Length()-iEndPos);
            iLastPos = 1;

        }
        else if ( asTmp == "/STRIKE")
        {
            Canvas->TextOut(tmpSize.cx+X, Y, asText.SubString(1, iStartPos-1));
            tmpSize.cx += Canvas->TextExtent(asText.SubString(1, iStartPos-1)).cx;

            // Set new font
            Canvas->Font->Style = Canvas->Font->Style >> fsStrikeOut;
            // Skip the <..> part
            asText = asText.SubString(iEndPos+1, asText.Length()-iEndPos);
            iLastPos = 1;
        }
        else
        {
            // Unsupported tag
            Canvas->TextOut(tmpSize.cx+X, Y, asText.SubString(1, iEndPos));
            tmpSize.cx += Canvas->TextExtent(asText.SubString(1, iEndPos)).cx;
            asText = asText.SubString(iEndPos+1, asText.Length()-iEndPos);
            iLastPos = 1;
        }

    } // end while(1)
}
//---------------------------------------------------------------------------
void __fastcall THintBox::SetHTMLFormatting(bool Value)
{
    if (Value != bHTMLFormatting)
    {
        bHTMLFormatting = Value;

        if (bHTMLFormatting)
            Parse();

        Paint();
    }
}
//---------------------------------------------------------------------------
void __fastcall THintBox::Update()
{
    if (bHTMLFormatting)
    {
        Parse();
    }

    Paint();
}
//---------------------------------------------------------------------------
void __fastcall THintBox::SetHintBox(bool Value)
{
    if (Value != bDrawBox)
    {
        bDrawBox = Value;

        Paint();
    }
}
//---------------------------------------------------------------------------

